package edu.berkeley.sbp;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.Sequence.Position;
import edu.berkeley.sbp.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;

/** a parser which translates streams of Tokens of type T into a Forest<R> */
public abstract class Parser<T extends Token, R> {

    public final Table pt;

    /** create a parser to parse the grammar with start symbol <tt>u</tt> */
    protected Parser(Union u)  { this.pt = new Table(u, top()); }
    protected Parser(Table pt) { this.pt = pt; }

    public abstract Forest<R> shiftedToken(T t, Token.Location loc);
    public abstract Topology<T> top();


    /** parse <tt>input</tt> for a exactly one unique result, throwing <tt>Ambiguous</tt> if not unique or <tt>Failed</tt> if none */
    public Tree<R> parse1(Token.Stream<T> input) throws IOException, Failed, Ambiguous {
        Forest<R> ret = parse(input);
        try { return ret.expand1(); }
        catch (Ambiguous a) {
            System.out.println("while expanding:");
            System.out.println(ret);
            throw a;
        }
    }

    /** parse <tt>input</tt>, using the table <tt>pt</tt> to drive the parser */
    public Forest<R> parse(Token.Stream<T> input) throws IOException, Failed {
        GSS gss = new GSS();
        Token.Location loc = input.getLocation();
        GSS.Phase current = gss.new Phase(null, this, null, input.next(1, 0, 0), loc, null);
        current.newNode(null, Forest.leaf(null, null), pt.start, true);
        int count = 1;
        for(;;) {
            loc = input.getLocation();
            current.reduce();
            Forest forest = current.token==null ? null : shiftedToken((T)current.token, loc);
            GSS.Phase next = gss.new Phase(current, this, current, input.next(count, gss.resets, gss.waits), loc, forest);
            count = next.hash.size();
            if (current.isDone()) return (Forest<R>)current.finalResult;
            current = next;
        }
    }
    

    // Exceptions //////////////////////////////////////////////////////////////////////////////

    public static class Failed extends RuntimeException {
        private final Token.Location location;
        private final String         message;
        public Failed() { this("", null); }
        public Failed(String message, Token.Location loc) { this.location = loc; this.message = message; }
        public Token.Location getLocation() { return location; }
        public String toString() { return message/* + (location==null ? "" : (" at " + location))*/; }
    }

    public static class Ambiguous extends RuntimeException {
        public final Forest ambiguity;
        public Ambiguous(Forest ambiguity) { this.ambiguity = ambiguity; }
        public String toString() {
            StringBuffer sb = new StringBuffer();
            sb.append("unresolved ambiguity "/*"at " + ambiguity.getLocation() + ":"*/);
            for(Object result : ambiguity.expand(false))
                sb.append("\n    " + result);
            return sb.toString();
        }
    }


    // Table //////////////////////////////////////////////////////////////////////////////

    /** an SLR(1) parse table which may contain conflicts */
    static class Table extends Walk.Cache {

        public final Walk.Cache cache = this;

        public HashMapBag<Position,State> byPosition = new HashMapBag<Position,State>();
        
        private void walk(Element e, HashSet<Element> hs) {
            if (e==null) return;
            if (hs.contains(e)) return;
            hs.add(e);
            if (e instanceof Atom) return;
            for(Sequence s : (Union)e) {
                hs.add(s);
                for(Position p = s.firstp(); p != null; p = p.next())
                    walk(p.element(), hs);
            }
        }

        /** the start state */
        public final State   start;

        /** used to generate unique values for State.idx */
        private int master_state_idx = 0;

        /** construct a parse table for the given grammar */
        public Table(Topology top) { this("s", top); }
        public Table(String startSymbol, Topology top) { this(new Union(startSymbol), top); }
        public Table(Union ux, Topology top) {
            Union start0 = new Union("0");
            start0.add(new Sequence.Singleton(ux, null, null));

            for(Sequence s : start0) cache.eof.put(s, true);
            cache.eof.put(start0, true);

            // construct the set of states
            HashMap<HashSet<Position>,State>   all_states    = new HashMap<HashSet<Position>,State>();
            HashSet<Element>                   all_elements  = new HashSet<Element>();
            walk(start0, all_elements);
            for(Element e : all_elements)
                cache.ys.put(e, new Walk.YieldSet(e, cache).walk());
            HashSet<Position> hp = new HashSet<Position>();
            reachable(start0, hp);
            this.start = new State(hp, all_states, all_elements);

            // for each state, fill in the corresponding "row" of the parse table
            for(State state : all_states.values())
                for(Position p : state.hs) {

                    // the Grammar's designated "last position" is the only accepting state
                    if (start0.contains(p.owner()) && p.next()==null)
                        state.accept = true;

                    if (p.isRightNullable(cache)) {
                        Walk.Follow wf = new Walk.Follow(top.empty(), p.owner(), all_elements, cache);
                        Reduction red = new Reduction(p);

                        Topology follow = wf.walk(p.owner());
                        if (p.owner() instanceof Sequence.RewritingSequence &&
                            (((Sequence.RewritingSequence)p.owner()).tag+"").equals("emailaddr")) {
                            System.out.println("follow before: " + new edu.berkeley.sbp.misc.CharToken.CharRange(follow));
                        }
                        for(Position p2 = p; p2 != null && p2.element() != null; p2 = p2.next())
                            follow = follow.intersect(new Walk.Follow(top.empty(), p2.element(), all_elements, cache).walk(p2.element()));
                        if (p.owner() instanceof Sequence.RewritingSequence &&
                            (((Sequence.RewritingSequence)p.owner()).tag+"").equals("emailaddr")) {
                            System.out.println("follow after: " + new edu.berkeley.sbp.misc.CharToken.CharRange(follow));
                        }
                        state.reductions.put(follow, red);
                        if (wf.includesEof()) state.eofReductions.add(red);
                    }

                    // if the element following this position is an atom, copy the corresponding
                    // set of rows out of the "master" goto table and into this state's shift table
                    if (p.element() != null && p.element() instanceof Atom)
                        state.shifts.addAll(state.gotoSetTerminals.subset(((Atom)p.element())));
                }
            for(State state : all_states.values()) {
                state.oreductions = state.reductions.optimize();
                state.oshifts = state.shifts.optimize();
            }
        }

        /** a single state in the LR table and the transitions possible from it */

        public class State implements Comparable<Table.State>, IntegerMappable, Iterable<Position> {
        
            public int toInt() { return idx; }

            public boolean lame() {
                for(Position p : this)
                    for(Position p2 = p; p2!=null; p2=p2.next())
                        if (p2.isLast() && !p2.owner().lame)
                            return false;
                return true;
            }
            /*
            public boolean isResolvable(Token t) {
                boolean found = false;
                for(Reduction r : getReductions(t)) {
                    Position p = r.position;
                    if (!p.isRightNullable(cache)) continue;
                    if (p.owner().firstp()==p) continue;
                    if (found) {
                        // found two items meeting criteria #1
                        return false;
                    } else {
                        found = true;
                        continue;
                    }
                    if (p.element()==null) continue;
                    Topology first = new Walk.First(top(), cache).walk(p.element());
                    if (first.contains(t))
                }
            }
            */

            public  final     int               idx    = master_state_idx++;
            private final     HashSet<Position> hs;

            private transient HashMap<Element,State>          gotoSetNonTerminals = new HashMap<Element,State>();
            private transient TopologicalBag<Token,State>     gotoSetTerminals    = new TopologicalBag<Token,State>();

            private           TopologicalBag<Token,Reduction> reductions          = new TopologicalBag<Token,Reduction>();
            private           HashSet<Reduction>              eofReductions       = new HashSet<Reduction>();
            private           TopologicalBag<Token,State>     shifts              = new TopologicalBag<Token,State>();
            private           boolean                         accept              = false;

            private VisitableMap<Token,State> oshifts = null;
            private VisitableMap<Token,Reduction> oreductions = null;

            // Interface Methods //////////////////////////////////////////////////////////////////////////////

            public boolean             isAccepting()               { return accept; }

            public boolean             canShift(Token t)           { return oshifts.contains(t); }
            public boolean             canReduce(Token t)          { return t==null ? eofReductions.size()>0 : oreductions.contains(t); }

            public Iterator<Position>  iterator()                  { return hs.iterator(); }

            public <B,C> void          invokeShifts(Token t, Invokable<State,B,C> irbc, B b, C c) {
                oshifts.invoke(t, irbc, b, c);
            }
            public <B,C> void          invokeReductions(Token t, Invokable<Reduction,B,C> irbc, B b, C c) {
                if (t==null) for(Reduction r : eofReductions) irbc.invoke(r, b, c);
                else         oreductions.invoke(t, irbc, b, c);
            }

            // Constructor //////////////////////////////////////////////////////////////////////////////

            /**
             *  create a new state consisting of all the <tt>Position</tt>s in <tt>hs</tt>
             *  @param hs           the set of <tt>Position</tt>s comprising this <tt>State</tt>
             *  @param all_states   the set of states already constructed (to avoid recreating states)
             *  @param all_elements the set of all elements (Atom instances need not be included)
             *  
             *   In principle these two steps could be merged, but they
             *   are written separately to highlight these two facts:
             * <ul>
             * <li> Non-atom elements either match all-or-nothing, and do not overlap
             *      with each other (at least not in the sense of which element corresponds
             *      to the last reduction performed).  Therefore, in order to make sure we
             *      wind up with the smallest number of states and shifts, we wait until
             *      we've figured out all the token-to-position multimappings before creating
             *      any new states
             *  
             * <li> In order to be able to run the state-construction algorithm in a single
             *      shot (rather than repeating until no new items appear in any state set),
             *      we need to use the "yields" semantics rather than the "produces" semantics
             *      for non-Atom Elements.
             *  </ul>
             */
            public State(HashSet<Position> hs,
                         HashMap<HashSet<Position>,State> all_states,
                         HashSet<Element> all_elements) {
                this.hs = hs;

                // register ourselves in the all_states hash so that no
                // two states are ever created with an identical position set
                all_states.put(hs, this);
                for(Position p : hs) byPosition.add(p,this);

                // Step 1a: examine all Position's in this state and compute the mappings from
                //          sets of follow tokens (tokens which could follow this position) to sets
                //          of _new_ positions (positions after shifting).  These mappings are
                //          collectively known as the _closure_

                TopologicalBag<Token,Position> bag0 = new TopologicalBag<Token,Position>();
                for(Position position : hs) {
                    if (position.isLast() || !(position.element() instanceof Atom)) continue;
                    Atom a = (Atom)position.element();
                    HashSet<Position> hp = new HashSet<Position>();
                    reachable(position.next(), hp);
                    bag0.addAll(a, hp);
                }

                // Step 1b: for each _minimal, contiguous_ set of characters having an identical next-position
                //          set, add that character set to the goto table (with the State corresponding to the
                //          computed next-position set).

                for(Topology<Token> r : bag0) {
                    HashSet<Position> h = new HashSet<Position>();
                    for(Position p : bag0.getAll(r)) h.add(p);
                    gotoSetTerminals.put(r, all_states.get(h) == null ? new State(h, all_states, all_elements) : all_states.get(h));
                }

                // Step 2: for every non-Atom element (ie every Element which has a corresponding reduction),
                //         compute the closure over every position in this set which is followed by a symbol
                //         which could yield the Element in question.
                //
                //         "yields" [in one or more step] is used instead of "produces" [in exactly one step]
                //         to avoid having to iteratively construct our set of States as shown in most
                //         expositions of the algorithm (ie "keep doing XYZ until things stop changing").
                /*
                for(Element e : all_elements) {
                    if (e instanceof Atom) continue;
                    HashSet<Position> h = new Walk.Closure(null, g.cache).closure(e, hs);
                    State s = all_states.get(h) == null ? new State(h, all_states, all_elements) : all_states.get(h);
                    if (gotoSetNonTerminals.get(e) != null)
                        throw new Error("this should not happen");
                    gotoSetNonTerminals.put(e, s);
                }
                */
                HashMapBag<Element,Position> move = new HashMapBag<Element,Position>();
                for(Position p : hs) {
                    Element e = p.element();
                    if (e==null) continue;
                    HashSet<Element> ys = cache.ys.get(e);
                    if (ys != null) {
                        for(Element y : ys) {
                            HashSet<Position> hp = new HashSet<Position>();
                            reachable(p.next(), hp);
                            move.addAll(y, hp);
                        }
                    }
                }
                for(Element y : move) {
                    HashSet<Position> h = move.getAll(y);
                    State s = all_states.get(h) == null ? new State(h, all_states, all_elements) : all_states.get(h);
                    gotoSetNonTerminals.put(y, s);
                }
            }

            public String toString() {
                //return "state["+idx+"]";
                StringBuffer ret = new StringBuffer();
                ret.append("state["+idx+"]: ");
                for(Position p : this) ret.append("{"+p+"}  ");
                return ret.toString();
            }

            public int compareTo(Table.State s) { return idx==s.idx ? 0 : idx < s.idx ? -1 : 1; }
        }

        /**
         *  the information needed to perform a reduction; copied here to
         *  avoid keeping references to <tt>Element</tt> objects in a Table
         */
        public class Reduction {
            // FIXME: cleanup; almost everything in here could go in either Sequence.Position.getRewrite() or else in GSS.Reduct
            public final int numPop;
            /*private*/ final Position position;
            private final Forest[] holder;    // to avoid constant reallocation
            public int hashCode() { return position.hashCode(); }
            public boolean equals(Object o) {
                if (o==null) return false;
                if (o==this) return true;
                if (!(o instanceof Reduction)) return false;
                Reduction r = (Reduction)o;
                return r.position == position;
            }
            public Reduction(Position p) {
                this.position = p;
                this.numPop = p.pos;
                this.holder = new Forest[numPop];
            }
            public String toString() { return "[reduce " + position + "]"; }

            private Forest zero = null;
            public Forest zero() {
                if (zero != null) return zero;
                if (numPop > 0) throw new Error();
                return zero = position.rewrite(null);
            }

            public void reduce(GSS.Phase.Node parent) {
                if (numPop==0) finish(parent, zero(), parent.phase());
                else           reduce(parent, numPop-1, parent.phase());
            }

            public void reduce(GSS.Phase.Node parent, GSS.Phase.Node onlychild) {
                if (numPop<=0) throw new Error("called wrong form of reduce()");
                int pos = numPop-1;
                Forest old = holder[pos];
                holder[pos] = parent.pending();
                if (pos==0) {
                    System.arraycopy(holder, 0, position.holder, 0, holder.length);
                    finish(onlychild, position.rewrite(parent.phase().getLocation()), parent.phase());
                } else {
                    reduce(onlychild, pos-1, parent.phase());
                }
                holder[pos] = old;
            }

            // FIXME: this could be more elegant and/or cleaner and/or somewhere else
            private void reduce(GSS.Phase.Node parent, int pos, GSS.Phase target) {
                Forest old = holder[pos];
                holder[pos] = parent.pending();
                if (pos==0) {
                    System.arraycopy(holder, 0, position.holder, 0, holder.length);
                    for(int i=0; i<position.pos; i++) if (position.holder[i]==null) throw new Error("realbad");
                    Forest rex = position.rewrite(target.getLocation());
                    for(GSS.Phase.Node child : parent.parents()) finish(child, rex, target);
                } else {
                    for(GSS.Phase.Node child : parent.parents()) reduce(child, pos-1, target);
                }
                holder[pos] = old;
            }
            private void finish(GSS.Phase.Node parent, Forest result, GSS.Phase target) {
                State state = parent.state.gotoSetNonTerminals.get(position.owner());
                if (result==null) throw new Error();
                if (state!=null)
                    target.newNode(parent, result, state, numPop<=0, this);
            }
        }
    }

    private static final Forest[] emptyForestArray = new Forest[0];


    // Helpers //////////////////////////////////////////////////////////////////////////////

    private static void reachable(Element e, HashSet<Position> h) {
        if (e instanceof Atom) return;
        for(Sequence s : ((Union)e))
            reachable(s.firstp(), h);
    }
    private static void reachable(Position p, HashSet<Position> h) {
        if (h.contains(p)) return;
        h.add(p);
        if (p.element() != null) reachable(p.element(), h);
    }

}
